from typing import Iterable, Iterator, List, Optional, Sized, Tuple, TypeVar, Union

import numpy as np
from torchdata.dataloader2 import DataLoader2
from torchdata.dataloader2.adapter import Adapter
from torchdata.dataloader2.reading_service import MultiProcessingReadingService
from torchdata.datapipes import functional_datapipe, DataChunk
from torchdata.datapipes.map import MapDataPipe
from torchdata.datapipes.iter import IterDataPipe

from pyPhasesML.DataAugmentation import DataAugmentation

T_co = TypeVar("T_co", covariant=True)
T = TypeVar("T")


# @functional_datapipe("batchFirst")
# class BatcherIterDataPipe(IterDataPipe[DataChunk]):
#     r"""
#     use the first dimension of the data as batch dimension
#     """
#     datapipe: IterDataPipe
#     batch_size: int
#     drop_last: bool

#     def __init__(
#         self,
#         datapipe: IterDataPipe,
#         batch_size: int,
#         count: int,
#         drop_last: bool = False,
#         wrapper_class=DataChunk,
#     ) -> None:
#         assert batch_size > 0, "Batch size is required to be larger than 0!"
#         super().__init__()
#         self.datapipe = datapipe
#         self.batch_size = batch_size
#         self.count = count
#         self.drop_last = drop_last
#         self.wrapper_class = wrapper_class

#     def __iter__(self) -> Iterator[DataChunk]:
#         batchX: List = []
#         batchY: List = []
#         for record in self.datapipe:
#             x, y = record
#             for i in range(len(x)):
#                 batchX.append(x[i])
#                 batchY.append(y[i])
#                 if len(batchX) == self.batch_size:
#                     yield self.wrapper_class((np.array(batchX), np.array(batchY)))
#                     batchX: List = []
#                     batchY: List = []
#         if len(batchX) > 0:
#             if not self.drop_last:
#                 yield self.wrapper_class((np.array(batchX), np.array(batchY)))

#     def __len__(self) -> int:
#         if self.drop_last:
#             return self.count // self.batch_size
#         else:
#             return (self.count + self.batch_size - 1) // self.batch_size

@functional_datapipe("batchFirst")
class BatcherMapDataPipe(MapDataPipe[DataChunk]):
    r"""
    batch the data by the first dimension
    """
    datapipe: MapDataPipe
    batch_size: int
    drop_last: bool

    def __init__(self,
                 datapipe: MapDataPipe[T],
                 batch_size: int,
                 drop_last: bool = False,
                 wrapper_class=DataChunk,
                 ) -> None:
        assert batch_size > 0, "Batch size is required to be larger than 0!"
        super().__init__()
        self.datapipe = datapipe
        self.batch_size = batch_size
        self.drop_last = drop_last
        self.wrapper_class = wrapper_class

    def __getitem__(self, index) -> DataChunk:
        batchX: List = []
        batchY: List = []
        indices = range(index * self.batch_size, (index + 1) * self.batch_size)
        try:
            for i in indices:
                X, Y = self.datapipe[i]
                if len(X) > 1:
                    raise Exception(f"Batches need to single dimension")
                batchX.append(X[0])
                batchY.append(Y[0])
            return self.wrapper_class((np.array(batchX), np.array(batchY)))
        except IndexError as e:
            if not self.drop_last and len(batchX) > 0:
                return self.wrapper_class((batchX, batchY))
            else:
                raise IndexError(f"Index {index} is out of bound.") from e

    def __len__(self) -> int:
        if isinstance(self.datapipe, Sized):
            if self.drop_last:
                return len(self.datapipe) // self.batch_size
            else:
                return (len(self.datapipe) + self.batch_size - 1) // self.batch_size
        else:
            raise TypeError("{} instance doesn't have valid length".format(type(self).__name__))


@functional_datapipe("unzipXY")
class ZipperMapDataPipe(MapDataPipe[Tuple[T_co, ...]]):
    datapipes: Tuple[MapDataPipe[T_co], ...]

    def __init__(self, datapipe: MapDataPipe[T_co]) -> None:
        self.datapipe = datapipe
        self.index = 0
        
    def __len__(self):
        return len(self.datapipe)

    def __getitem__(self, index) -> Tuple[T_co, ...]:
        x, y = zip(*self.datapipe[index])
        return np.array(x), np.array(y)


# @functional_datapipe("augment")
# class Augmentor(IterDataPipe):
#     datapipes: IterDataPipe

#     def __init__(self, datapipe: MapDataPipe[T_co], augmentation: DataAugmentation) -> None:
#         self.datapipe = datapipe
#         self.augmentation = augmentation

#     def __len__(self):
#         return len(self.datapipe)

#     def __iter__(self) -> int:
#         for r in self.datapipe:
#             yield self.augmentation(r)


@functional_datapipe("augment")
class Augmentor(MapDataPipe):
    datapipes: MapDataPipe

    def __init__(self, datapipe: MapDataPipe[T_co], augmentation: DataAugmentation, config) -> None:
        self.datapipe = datapipe
        self.augmentation = augmentation
        self.config = config

    def __getitem__(self, index) -> MapDataPipe[T_co]:
        return self.augmentation(self.datapipe[index], self.config)

    def __len__(self) -> MapDataPipe[T_co]:
        return len(self.datapipe)


@functional_datapipe("segmentWise")
class Segments(MapDataPipe):
    datapipes: MapDataPipe

    def __init__(self, datapipe: MapDataPipe[T_co], recordLengths, segmentLength) -> None:
        self.datapipe = datapipe
        self.segmentLength = segmentLength
        
        # shape = self.datapipe.dataExporterSignals.fileShape
        recordLengths = np.array(recordLengths)
        recordSegmentLength = recordLengths // segmentLength
        recordSegmentLengthIndex = np.insert(recordSegmentLength.cumsum(), 0, 0)
        segmentCount = sum(recordLengths) // segmentLength

        recordMaping = np.concatenate([np.array([i] * length) for i, length in enumerate(recordSegmentLength)])
        segmentIndexes = np.arange(segmentCount) - recordSegmentLengthIndex[recordMaping]
        # mapping returns for each segment the record and the segment index
        self.segmentMapping = np.hstack([recordMaping.reshape(-1, 1), segmentIndexes.reshape(-1, 1)])
        self.segmentCount = segmentCount

    def __len__(self):
        return self.segmentCount

    def __getitem__(self, index) -> Tuple[T_co, ...]:
        recordIndex, segmentIndex = self.segmentMapping[index]
        x, y = self.datapipe[recordIndex]
        x = x.reshape(-1, self.segmentLength, x.shape[-1])
        y = y.reshape(x.shape[0], -1, y.shape[-1])
        return x[segmentIndex], y[segmentIndex]


class AugmentationDataset(MapDataPipe):
    def __init__(self, dataExporterSignals, dataExporterFeatures):
        self.dataExporterSignals = dataExporterSignals
        self.dataExporterFeatures = dataExporterFeatures
        self.segmentMapping = None

    def __len__(self):
        return len(self.dataExporterSignals)

    def __getitem__(self, idx):
        segmentX, segmentY = self.dataExporterSignals[idx], self.dataExporterFeatures[idx]
        # convert to numpy arrays
        return np.array(segmentX), np.array(segmentY)

    def __iter__(self) -> int:
        for d in self:
            yield d


class AugmentationDataLoaderXY(DataLoader2):
    def __init__(
        self,
        datapipe: MapDataPipe,
        threads: int = None,
        pinMemory: bool = False,
        datapipe_adapter_fn: Optional[Union[Iterable[Adapter], Adapter]] = None,
    ) -> None:
        reading_service = None
        if threads is not None:
            reading_service = MultiProcessingReadingService(num_workers=threads, pin_memory=pinMemory)
        super().__init__(datapipe, datapipe_adapter_fn, reading_service)

    def __len__(self):
        return len(self.datapipe)

    def generator(self, wrapper=None):
        while True:
            for d in self:
                yield d if wrapper is None else wrapper(d)

    def __getitem__(self, idx):
        return self.datapipe[idx]



import math
from collections.abc import Iterable
from dataclasses import astuple, dataclass

import numpy as np


class ValidDataset:
    pass


@dataclass
class DataSet:
    x: np.array
    y: np.array
    numClasses: int = None
    catMatrix = None

    def __len__(self):
        return 2

    @staticmethod
    def fromTuple(data, numClasses=None):
        x, y = data
        return DataSet(x, y, numClasses)

    def asTuple(self):
        return astuple(self)

    def __getitem__(self, index):
        if index == 0:
            return self.x
        if index == 1:
            return self.y
        if index == 2:
            raise StopIteration
        raise TypeError("Datset can only be split in x and y, try to get index %i" % (index))


class DataSetContinous(ValidDataset):
    def __len__(self):
        return self.recordCount if self.recordWise else self.batchCount

    def __init__(
        self,
        batchFunction,
        batchCount=None,
        recordCount=None,
        numClasses=None,
        getSingleRecord=None,
    ) -> None:
        self.batchFunction = batchFunction
        self.batchCount = batchCount
        self.recordCount = recordCount
        self.numClasses = numClasses
        self.continous = True
        self.index = 0
        self.recordWise = False
        self.getSingleRecord = getSingleRecord
        self.augmentation = True
        self.function = None
        self.distinct = True
        if self.recordWise and self.recordCount is None:
            self.recordCount = len(batchFunction)
        if not self.recordWise and self.batchCount is None:
            self.batchCount = len(batchFunction)

    def getFunction(self):
        if self.function is None:
            if callable(self.batchFunction):
                self.function = self.batchFunction(recordWise=self.recordWise, augmentation=self.augmentation)
            else:
                self.function = iter(self.batchFunction)
        return self.function

    def nonContinous(self):
        self.continous = False

    def __next__(self):
        if not self.continous and self.index == len(self):
            raise StopIteration

        dataTuple = self.getFunction().__next__()
        batch = DataSet.fromTuple(dataTuple, numClasses=self.numClasses)
        self.index += 1
        return batch

    def __iter__(self):
        self.function = None
        self.getFunction().__iter__()
        self.index = 0
        return self

    def generator(self, wrapper=None):
        while True:
            ret = self.__next__()
            wrapper = (lambda x: x) if wrapper is None else wrapper
            yield wrapper(ret)

    def __getitem__(self, index):
        dataTuple = self.getSingleRecord(index)
        batch = DataSet.fromTuple(dataTuple, numClasses=self.numClasses)
        return batch


class TrainingSetLoader:
    def __init__(self, trainingData=None, validationData=None):
        self.trainingData: DataSet = trainingData
        self.validationData: DataSet = validationData


class TrainingSet:
    def __init__(
        self,
        trainingData=None,
        validationData=None,
        numClasses=None,
        useContinous=True,
        distinct=True,
    ) -> None:

        self.trainingData: DataSet = None
        self.validationData: DataSet = None
        self.transformChannelLast = False

        if isinstance(trainingData, ValidDataset):
            self.trainingData = trainingData
        elif isinstance(trainingData, Iterable) and useContinous:
            self.trainingData = DataSetContinous(trainingData)
        else:
            self.trainingData = DataSet.fromTuple(trainingData, len(trainingData))

        if isinstance(validationData, ValidDataset):
            self.validationData = validationData
        elif isinstance(validationData, Iterable) and useContinous:
            self.validationData = DataSetContinous(validationData, len(validationData))
        else:
            self.validationData = DataSet.fromTuple(validationData)

        self.trainingData.distinct = distinct
        self.validationData.distinct = distinct

        self.trainingData.numClasses = numClasses
        self.validationData.numClasses = numClasses

    def toCategorical(self):
        self.trainingData.toCategorical()
        self.validationData.toCategorical()

    def getNextTrainingBatch(self):
        dataTuple = self.trainingData.__next__()
        batch = DataSet.fromTuple(dataTuple)
        return batch


class DatasetWrap(DataSetContinous):
    def __getitem__(self, index):
        return self.dataset[index]

    def __len__(self):
        return len(self.dataset)

    def __init__(self, dataset) -> None:
        self.dataset = dataset

    def nonContinous(self):
        self.continous = False

    def __next__(self):
        return self.dataset.__next__()

    def __iter__(self):
        return self.dataset.__iter__()

    def asTuple(self):
        return self


class DatasetWrapXY(ValidDataset):
    def __init__(self, dataset, batchSize=1) -> None:
        self.x, self.y = dataset
        self.batchSize = batchSize
        self.length = math.ceil(len(self.x) / self.batchSize)

    def __getitem__(self, index):
        if index >= self.length:
            raise IndexError
        s = slice(index * self.batchSize, (index + 1) * self.batchSize)
        return self.x[s], self.y[s]

    def __len__(self):
        return self.length

    def nonContinous(self):
        self.continous = False

    def generator(self, wrapper=None):
        while True:
            for d in self:
                wrapper = (lambda x: x) if wrapper is None else wrapper
                yield wrapper(DataSet.fromTuple(d))


class DatasetWrapItemGenerator(ValidDataset):
    def __init__(self, generator, lengths, batchSize=1) -> None:
        self.generator = generator
        self.length = lengths
        self.batchSize = batchSize

    def __len__(self):
        return self.length

    def nonContinous(self):
        self.continous = False

    def generator(self, wrapper=None):
        wrapper = (lambda x: x) if wrapper is None else wrapper
        yield wrapper(self.generator.__next__())

        batch = []
        curRecordsInBatch = 0

        while True:
            r = self.generator.__next__()
            batch.append(r)
            curRecordsInBatch += 1
            if curRecordsInBatch == self.batchSize:
                yield r
                batch = []
                curRecordsInBatch = 0


class DatasetWrapBatchGenerator(ValidDataset):
    def __init__(self, generator, length) -> None:
        self.generatorFunction = generator
        self.length = length
        self.index = 0

    def __len__(self):
        return self.length

    def generator(self, wrapper=None):
        if wrapper is None:
            while 1:
                yield next(self.generatorFunction)
        else:
            while 1:
                yield wrapper(next(self.generatorFunction))

    def __iter__(self):
        self.generatorFunction.__iter__()
        self.index = 0
        return self

    def __next__(self):
        if self.index == self.length:
            raise StopIteration
        self.index += 1
        return next(self.generatorFunction)



from numpy.core.records import record
import tensorflow as tf
import keras as K
import numpy as np

from ..DataSet import TrainingSet
from ..Model import Model


# K.set_image_data_format("channels_last")


class ModelTf1Adapter(Model):
    model = None

    def _metrics(self, metric, validation=False):
        prepend = "val_" if validation else ""
        return prepend + metric["name"]

    def beforeTrain(self, dataset):
        # from tensorflow.keras.models import load_model
        if "loadModelPath" in self.options:
            loadedModel = tf.keras.models.load_model(self.options["loadModelPath"])
            try:
                self.model.set_weights(loadedModel.get_weights())
            except Exception:
                raise Exception
        if hasattr(self, "freeze"):
            layers_freeze = self.freeze
            # len(self.model.layers)
            for layer in self.model.layers[:layers_freeze]:
                layer.trainable = False
        return

    def train(self, dataset: TrainingSet):
        callbacks = []
        mainMetric = self.validationMetrics[0]
        metricStop = self.getMetric(mainMetric)
        metricSave = self.getMetric(mainMetric)

        # tensorboard_callback = K.callbacks.TensorBoard(
        #     log_dir=self.logPath + "/tb",
        #     histogram_freq=1,
        # )
        # callbacks.append(tensorboard_callback)

        if self.stopAfterNotImproving is not None:
            earlyStopping_cb = K.callbacks.EarlyStopping(
                monitor=self._metrics(metricStop, True),
                patience=self.stopAfterNotImproving,
                verbose=0,
                mode=metricStop["type"],
                # baseline=None,
                # restore_best_weights=True,
            )
            callbacks.append(earlyStopping_cb)

        if self.learningRateDecay is not None:
            lr_decay_cb = K.callbacks.LearningRateScheduler(
                schedule=lambda epoch: self.learningRate * (self.learningRateDecay ** epoch)
            )
            callbacks.append(lr_decay_cb)

        log = K.callbacks.CSVLogger(self.getCsvPath())
        callbacks.append(log)

        checkpoint = K.callbacks.ModelCheckpoint(
            self.logPath + "/weights-{epoch:02d}.h5",
            monitor=self._metrics(metricSave, True),
            mode=metricSave["type"],
            save_best_only=True,
            save_weights_only=True,
            verbose=0,
        )
        callbacks.append(checkpoint)
        classWeight = None if self.classWeights is None else dict(list(zip(range(6), self.classWeights)))

        batchesPerEpoch = len(dataset.trainingData)
        validationBatchesPerEpoch = len(dataset.validationData)
        # train = dataset.trainingData.generator(wrapper=self.prepareData)
        # validationData = dataset.validationData.generator(wrapper=self.prepareData)

        def combine(dataset):
            dataset.continous = False
            recordsX = []
            recordsY = []
            for r in dataset:
                recordsX.append(r.x)
                recordsY.append(r.y)

            recordsX = np.concatenate(recordsX)
            recordsY = np.concatenate(recordsY)

            return recordsX, recordsY

        trainX, trainY = combine(dataset.trainingData)
        valX, valY = combine(dataset.validationData)

        self.cleanCsv()
        self.model.fit(
            [trainX, trainY],
            [trainY, trainX],  # capsnet specific
            validation_data=[[valX, valY], [valY, valX]],  # capsnet specific
            # steps_per_epoch=batchesPerEpoch,
            # validation_steps=validationBatchesPerEpoch,
            class_weight=classWeight,
            batch_size=self.batchSize,
            shuffle=False,
            verbose=1,
            epochs=self.maxEpochs,
            callbacks=callbacks,
        )

        metricHistory = self.model.history.history["val_capsnet_acc"]  # capsnet specific
        self.bestMetric = max(metricHistory)
        self.fullEpochs = len(metricHistory)
        return self.model

    def eval(model, data):
        x_test, y_test = data
        y_pred, x_recon = model.predict(x_test, batch_size=64)
        print("-" * 50)
        print(
            "Test acc:",
            np.sum(np.argmax(y_pred, 1) == np.argmax(y_test, 1)) / y_test.shape[0],
        )
        import matplotlib.pyplot as plt

        plt.figure()
        plt.subplot(211)
        plt.plot(x_recon[16], "b-")
        plt.title("Reconstruction")

        plt.figure()
        plt.subplot(212)
        plt.plot(x_test[16])
        plt.title("Original")

    def build(self):
        optimizer = None
        lossFunction = self.getLossFunction()
        lossWeights = self.getLossWeights()
        metrics = []

        if self.optimizer == "adams":
            optimizer = K.optimizers.Adam(lr=self.learningRate)

        for metric in self.monitorMetrics:
            metrics.append(metric)

        self.model.compile(
            optimizer=optimizer,
            loss=lossFunction,
            loss_weights=lossWeights,
            metrics=metrics,
            # run_eagerly=None,  # for debug purpose!
        )

    def summary(self):
        self.model.summary()
        # total_params = tf.reduce_sum([tf.reduce_prod(v.shape) for v in self.model.trainable_weightstf.trainable_variables()])
        total_params = 2134528

        self.log("Total trainable Parameters: %i" % (total_params))
        self.parameter = int(total_params)

        return self.model.summary()

    def __init__(self, options) -> None:
        super().__init__(options=options)

        # if self.useGPU:
        #     physical_devices = tf.config.experimental.list_physical_devices("GPU")
        #     print("Num GPUs Available: ", len(physical_devices))
        #     if len(physical_devices) > 0:
        #         tf.config.experimental.set_memory_growth(physical_devices[0], True)

    def save(self, path):
        self.model.save_weights(path)

    def load(self, path):
        return self.model.load_weights(path)

    def loadState(self, state):
        return self.model.load_weights(state)

    def predict(self, input, get_likelihood=False, returnNumpy=False):
        batchSize = input.shape[0]

        input = self.prepareX(input)

        if list(input.shape[1:]) != self.inputShape:
            input = input.reshape(-1, self.inputShape[0], self.inputShape[1], self.inputShape[2])
        input_tensor = tf.convert_to_tensor(input)
        output_tensor = self.model(input_tensor)
        out = output_tensor.numpy()
        out = out.reshape(1, out.shape[0] * out.shape[1], out.shape[2])

        predictions = self.mapOutputForPrediction(out)
        predictions = predictions.reshape(batchSize, -1, self.numClasses)

        if not get_likelihood:
            predictions = predictions.argmax(axis=2)

        # predict supposed to return Tensors: https://js.tensorflow.org/api/latest/#tf.Sequential.predict
        # but its allready numpy
        # if returnNumpy:
        #     predictions = predictions.numpy()

        return predictions



import importlib
import os
from pathlib import Path

from pyPhases.Data import DataNotFound
from pyPhases.exporter.DataExporter import DataExporter
from pyPhases.util.Logger import Logger

from ..Model import Model

if importlib.util.find_spec("tensorflow"):

    if importlib.util.find_spec("torch"):
        print(
            "\033[31;1;4m%s\033[0m"
            % "Tensorflow and PyTorch are installed in the same enviroment, only one model (Tensorflow) can be handled"
        )

    if importlib.util.find_spec("tensorflow.keras"):
        from tensorflow.keras import models
        from tensorflow.python.keras.engine.functional import Functional

    elif importlib.util.find_spec("keras"):
        from keras import models

        Functional = False

    class ModelExporter(DataExporter):
        includesStorage = True

        def initialOptions(self):

            return {"basePath": "data/"}

        def getPath(self, dataId):
            return self.getOption("basePath") + dataId

        def checkType(self, expectedType):
            funcCheck = issubclass(expectedType, Functional) if Functional is not None else False
            return issubclass(expectedType, models.Sequential) or issubclass(expectedType, Model) or funcCheck

        def read(self, dataId, options={}):
            path = self.getPath(dataId)
            if not Path("%s.index" % path).exists():
                raise DataNotFound()
            return path

        def write(self, dataId, model, options={}):
            model.save_weights(self.getPath(dataId))

elif importlib.util.find_spec("torch"):
    import torch

    class ModelExporter(DataExporter):
        includesStorage = True

        def initialOptions(self):

            return {"basePath": "data/"}

        def getPath(self, dataId):
            return self.getOption("basePath") + dataId

        def checkType(self, expectedType):

            return issubclass(expectedType, torch.nn.Module) or issubclass(expectedType, Model)

        def read(self, dataId, options={}):
            try:
                deviceName = "cuda" if torch.cuda.is_available() else "cpu"
                return torch.load(self.getPath(dataId), map_location=torch.device(deviceName))
            except FileNotFoundError:
                raise DataNotFound()

        def write(self, dataId, model, options={}):
            torch.save(model.state_dict(), self.getPath(dataId))

else:
    # ability to ignore missing model exporter for testing purposes
    if os.environ.get("PYPHASESML_IGNORE_MISSING_MODEL_EXPORTER", 0) == 1:
        raise Exception("No supported ModelExporter located (Supported: pytorch/tensorflow)")
    else:

        class ModelExporter(DataExporter):
            includesStorage = True

            def checkType(self, expectedType):
                return False

            def read(self, path, data, options={}):
                raise Exception("ModelExporter is only as stub, please install tensorflow or pytorch to use the ModelExporter")

            def write(self, path, data, model, options={}):
                raise Exception("ModelExporter is only as stub, please install tensorflow or pytorch to use the ModelExporter")



import importlib

from pyPhases import ConfigNotFoundException

from .Model import Model


class ModelManager:
    model: Model = None
    modelOptions = None
    beforeBuild = None
    modelPath = None

    @staticmethod
    def getModel(forceReload=False) -> Model:
        if ModelManager.model is None or forceReload:
            if ModelManager.modelOptions is None:
                raise Exception("ModelManager setModel was never called!")
            else:
                ModelManager.loadModelFromOptions()
        return ModelManager.model

    def loadModelFromOptions() -> None:

        options = ModelManager.modelOptions
        modelClass = ModelManager.loadModelByModule(options["modelName"])
        model = modelClass(options["modelConfig"])

        for field in options["general"]:
            setattr(model, field, options["general"][field])

        if ModelManager.beforeBuild is not None:
            ModelManager.beforeBuild(model)

        model.init()
        model.define()
        model.build()
        ModelManager.model = model

    def checkValueInConfig(self, config, value):

        if value not in config:
            raise ConfigNotFoundException("The value '%s' is required in the config" % value)

    def validate(config):
        def checkValueInConfig(config, value, valuePath=None):

            checkDict = config if valuePath is None else config[valuePath]

            if value not in checkDict:
                valuePath = value if valuePath is None else valuePath + "." + value
                raise ConfigNotFoundException("The value '%s' is required in the config" % valuePath)

        checkValueInConfig(config, "modelName")
        checkValueInConfig(config, "classification")
        checkValueInConfig(config, "trainingParameter")
        checkValueInConfig(config, "inputShape")
        checkValueInConfig(config, "classNames", "classification")
        checkValueInConfig(config, "batchSize", "trainingParameter")

        if not isinstance(config["classification"]["classNames"], list):
            raise ConfigNotFoundException("The value 'classification.classNames' is required to be a list of class names")

    @staticmethod
    def loadModel(project) -> None:
        config = project.config
        ModelManager.validate(config)
        ModelManager.modelPath = config["modelPath"]

        trainingParameter = config["trainingParameter"]
        classNames = config["classification"]["classNames"]
        classCount = len(classNames)
        ModelManager.modelOptions = {
            "modelName": project.config["modelName"],
            "modelConfig": project.getConfig("model", {}),
            "general": {
                "inputShape": project.config["inputShape"],
                "numClasses": classCount,
                "maxEpochs": project.getConfig("trainingParameter.maxEpochs", None, False),
                "batchSize": trainingParameter["batchSize"],
                "classWeights": project.getConfig("trainingParameter.classWeights", None, False),
                "classNames": classNames,
                "learningRate": project.getConfig("trainingParameter.learningRate", 0.001),
                "learningRateDecay": project.getConfig("trainingParameter.learningRateDecay", None, False),
                "stopAfterNotImproving": project.getConfig("trainingParameter.stopAfterNotImproving", 10),
                "validationEvery": project.getConfig("trainingParameter.validationEvery", None, False),
                "optimizer": project.getConfig("trainingParameter.optimizer", "Adams"),
                "ignoreClassIndex": project.getConfig("ignoreClassIndex", None, False),
                "validationMetrics": project.getConfig("trainingParameter.validationMetrics", ["loss"]),
                "useEventScorer": project.getConfig("trainingParameter.useEventScorer", False),
                "predictionType": project.getConfig("classification.type", "classification"),
                "oneHotDecoded": project.getConfig("oneHotDecoded", False),
            },
        }

    @staticmethod
    def loadModelByModule(name):
        path = ModelManager.modelPath.replace("/", ".")
        packageSplit = path.split(".")
        package = packageSplit[0]
        path = ".".join(packageSplit[1:])
        path = "." + path if path != "" else ""
        module = importlib.import_module("%s.%s.%s" % (path, name, name), package)
        # module = importlib.import_module(".%s.%s" % (name, name), package=userModels.__package__)
        return getattr(module, name)



from pathlib import Path
import shutil

import numpy as np
from torchdata.datapipes.iter import IterDataPipe
from torchdata.datapipes.map import MapDataPipe

from pyPhases.Data import DataNotFound
from pyPhases.exporter.DataExporter import DataExporter


class MemmapRecordExporter(DataExporter, IterDataPipe, MapDataPipe):
    includesStorage = True
    batchFillProcess = None

    def __init__(self, options=None):
        super().__init__(options)
        self.currentArrayShapes = {}

    def initialOptions(self):
        return {
            "basePath": "data/",
            "dtype": "float32",
        }

    def getShapeFilePath(self, dataId, tmp=False):
        add = "-tmp" if tmp else ""
        return self.getPath(dataId) + add + "-shape.npy"

    def saveShape(self, dataId, lengths, saveShape, tmp=False):
        shape = np.array((lengths, saveShape), dtype="object")
        np.save(self.getShapeFilePath(dataId, tmp=tmp), shape)
        self.log("saved %s: %s" % (dataId, str(shape)))

    def finishShape(self, dataId):
        self.saveShape(dataId, *self.currentArrayShapes[dataId], tmp=False)
        del self.currentArrayShapes[dataId]
        tmpPath = Path(self.getShapeFilePath(dataId, tmp=True))
        if tmpPath.is_file():
            tmpPath.unlink()

    def finishStream(self, dataId, options={}):
        if dataId in self.currentArrayShapes:
            self.finishShape(dataId)
        path = self.getPath(dataId)
        return shutil.move(path + "-tmp", path)

    def saveFromArrayList(self, dataId, arrayList):
        """fills the array from a list of 2-dimensional arrays"""
        array = np.concatenate(arrayList)
        array = np.expand_dims(array, 0)
        memmap = self.stream(dataId, {"dtype": "float32", "mode": "w+", "shape": array.shape})
        memmap[:, :, :] = array
        memmap.flush()
        del memmap

        lengths = [r.shape[0] for r in arrayList]
        self.currentArrayShapes[dataId] = lengths, array.shape
        self.finishStream(dataId)
        
    def loadTmp(self, dataId):
        shapePath = self.getShapeFilePath(dataId, tmp=True)
        if not Path(shapePath).exists():
            raise Exception("Shape file not found: %s" % shapePath)
        self.currentArrayShapes[dataId] = np.load(shapePath, allow_pickle=True)

    def saveAndAppendArray(self, dataId, arrayList):
        """appends an array to the existing file or creates a new one"""

        array = np.concatenate(arrayList)
        array = np.expand_dims(array, 0)

        memmap = None
        if dataId not in self.currentArrayShapes:
            self.currentArrayShapes[dataId] = [[], np.array((1, 0, array.shape[2]))]
            memmap = self.stream(dataId, {"dtype": "float32", "mode": "w+", "shape": array.shape})

        shape = self.currentArrayShapes[dataId][1]
        if array.shape[2] != shape[2]:
            raise Exception("The number of channels has to be constant, but was %s and is now %s" % (shape[2], array.shape[2]))

        lengths = shape[1]

        shape[1] += array.shape[1]

        if memmap is None:
            memmap = self.stream(
                dataId,
                {
                    "dtype": "float32",
                    "mode": "readwrite",
                    "shape": tuple(shape.tolist()),
                },
            )
        memmap[:, lengths:, :] = array
        memmap.flush()
        del memmap

        self.currentArrayShapes[dataId][0] += [r.shape[0] for r in arrayList]
        self.currentArrayShapes[dataId][1] = shape
        self.saveShape(dataId, *self.currentArrayShapes[dataId], tmp=True)

    def checkType(self, type):
        return type == np.memmap or type == IterDataPipe or type == MapDataPipe

    def getPath(self, dataId):
        return self.getOption("basePath") + dataId

    def stream(self, dataId, options):
        return np.memmap(self.getPath(dataId + "-tmp"), **options)

    def exists(self, dataId, options={}):
        return Path(self.getPath(dataId)).is_file()
    
    def existsTmp(self, dataId):
        return Path(self.getPath(dataId + "-tmp")).is_file()

    def read(self, dataId, options):
        for f, k in self.options.items():
            options.setdefault(f, k)

        if not self.exists(dataId):
            raise DataNotFound("Data with id %s nof found" % dataId)

        shapePath = self.getShapeFilePath(dataId)
        if not Path(shapePath).exists():
            raise Exception("Shape file not found: %s" % shapePath)

        # load complete shape, and record lengths
        lengths, shape = np.load(shapePath, allow_pickle=True)
        self.recordLengths = lengths
        self.fileShape = tuple(shape)
        self.type = options["dtype"]
        self.CurrentItemIndex = 0

        self.filePath = self.getPath(dataId)

        return self

    def __iter__(self):
        self.CurrentItemIndex = 0
        return self

    def __next__(self):
        if self.CurrentItemIndex >= len(self):
            raise StopIteration
        else:
            self.CurrentItemIndex += 1
            return self[self.CurrentItemIndex - 1]

    def __getitem__(self, index):
        memmap = np.memmap(self.filePath, dtype=self.type, mode="r", shape=self.fileShape)
        start = sum(self.recordLengths[:index])
        end = start + self.recordLengths[index]
        return memmap[0, start:end, :]

    def __len__(self):
        return len(self.recordLengths)



from ..scorer.ScorerTF import ScorerTF

import shutil
import tensorflow as tf
import tensorflow.keras as K
import matplotlib.pyplot as plt
import numpy as np
import io

from ..DataSet import TrainingSet
from ..Model import Model


def plot_to_image(figure):
    """Converts the matplotlib plot specified by 'figure' to a PNG image and
    returns it. The supplied figure is closed and inaccessible after this call."""
    # Save the plot to a PNG in memory.
    buf = io.BytesIO()
    plt.savefig(buf, format="png")
    # Closing the figure prevents it from being displayed directly inside
    # the notebook.
    plt.close(figure)
    buf.seek(0)
    # Convert PNG buffer to TF image
    image = tf.image.decode_png(buf.getvalue(), channels=4)
    # Add the batch dimension
    image = tf.expand_dims(image, 0)
    return image


class ModelKerasAdapter(Model):
    model: K.Model = None

    def _metrics(self, metric, validation=False):
        prepend = "val_" if validation else ""
        return prepend + metric["name"]

    def getMetrics(self):
        scorer = ScorerTF(self.numClasses)
        metrics = []
        for metric in self.validationMetrics:
            m = scorer.getMetricScorer(metric)
            metrics.append(m)
        return metrics

    def beforeTrain(self, dataset):
        # from tensorflow.keras.models import load_model
        if "loadModelPath" in self.options:
            loadedModel = tf.keras.models.load_model(self.options["loadModelPath"])
            try:
                self.model.set_weights(loadedModel.get_weights())
            except Exception:
                raise Exception
        if hasattr(self, "freeze"):
            layers_freeze = self.freeze
            # len(self.model.layers)
            for layer in self.model.layers[:layers_freeze]:
                layer.trainable = False
        return

    def train(self, dataset: TrainingSet):
        callbacks = []
        mainMetric = self._metrics(self.metrics[self.validationMetrics[0]])
        metricStop = self.getMetric(mainMetric)
        metricSave = self.getMetric(mainMetric)

        tbLogDir = self.logPath + "/tb"
        shutil.rmtree(tbLogDir, ignore_errors=True)

        # tf.debugging.experimental.enable_dump_debug_info(tbLogDir, tensor_debug_mode="FULL_HEALTH", circular_buffer_size=-1)
        tensorboard_callback = K.callbacks.TensorBoard(
            log_dir=tbLogDir,
            histogram_freq=1,
        )
        callbacks.append(tensorboard_callback)

        if self.predictionType == "reconstruction":
            model = self.model
            reconstruction_train = dataset.trainingData.__next__().x
            reconstruction_val = dataset.validationData.__next__().x
            reconstructionExamples = np.concatenate((reconstruction_train[0:5], reconstruction_val[0:5]))
            dataset.trainingData.__iter__()
            dataset.validationData.__iter__()

            file_writer_recon = tf.summary.create_file_writer(tbLogDir + "/recon")

            def log_examples(epoch, logs):
                reconstructions = model.predict(reconstructionExamples)
                for i, recon in enumerate(reconstructions):
                    type = "train" if i < 5 else "validation"
                    fig, ax = plt.subplots(1, figsize=(6, 3), dpi=300)
                    ax.plot(reconstructionExamples[i][:, 0])
                    ax.plot(recon[:, 0])
                    image = plot_to_image(fig)
                    with file_writer_recon.as_default():
                        tf.summary.image(
                            "Example Reconstruction (%s) %i" % (type, i),
                            image,
                            step=epoch,
                        )

            cm_callback = K.callbacks.LambdaCallback(on_epoch_end=log_examples)
            callbacks.append(cm_callback)

        if self.stopAfterNotImproving is not None:
            earlyStopping_cb = K.callbacks.EarlyStopping(
                monitor=self._metrics(metricStop, True),
                patience=self.stopAfterNotImproving,
                verbose=0,
                mode=metricStop["type"],
                baseline=None,
                restore_best_weights=True,
            )
            callbacks.append(earlyStopping_cb)

        if self.learningRateDecay is not None and self.learningRateDecay > 0:
            lr_decay_cb = K.callbacks.LearningRateScheduler(
                schedule=lambda epoch: self.learningRate * (self.learningRateDecay**epoch)
            )
            callbacks.append(lr_decay_cb)

        log = K.callbacks.CSVLogger(self.getCsvPath())
        callbacks.append(log)

        checkpoint = K.callbacks.ModelCheckpoint(
            self.logPath + "/weights-{epoch:02d}.h5",
            monitor=self._metrics(metricSave, True),
            mode=metricSave["type"],
            save_best_only=True,
            save_weights_only=True,
            verbose=0,
        )
        callbacks.append(checkpoint)
        classWeight = None if self.classWeights is None else dict(list(zip(range(6), self.classWeights)))

        dataset.trainingData.continous = True
        dataset.validationData.continous = True
        batchesPerEpoch = len(dataset.trainingData)
        validationBatchesPerEpoch = len(dataset.validationData)
        train = dataset.trainingData.generator(wrapper=self.prepareData)
        validationData = dataset.validationData.generator(wrapper=self.prepareData)

        self.cleanCsv()

        # debug
        # Model for trainingsdata: self.model(train.__next__()[0])
        # Model for validationdata: self.model(validationData.__next__()[0])
        # Loss (if its not implemented as string!): self.getLossFunction()(train.__next__()[1], self.model(train.__next__()[0]))
        # Test other Loss: tf.keras.losses.BinaryCrossentropy(from_logits=False)(train.__next__()[1], self.model(train.__next__()[0]))
        # Test Test Scorer
        # x, y = train.__next__()
        # ScorerTF(self.numClasses).getMetricScorer("kappa")(y, self.model(x))
        self.model.fit(
            x=train,
            validation_data=validationData,
            steps_per_epoch=batchesPerEpoch,
            validation_steps=validationBatchesPerEpoch,
            class_weight=classWeight,
            batch_size=self.batchSize,
            shuffle=False,
            verbose=1,
            epochs=self.maxEpochs,
            callbacks=callbacks,
        )

        metricHistory = self.model.history.history[mainMetric]
        self.bestMetric = max(metricHistory)
        self.fullEpochs = len(metricHistory)
        return self.model

    def build(self):
        optimizer = None
        lossFunction = self.getLossFunction()
        lossWeights = self.getLossWeights()

        if self.optimizer == "adams":
            optimizer = K.optimizers.Adam(learning_rate=self.learningRate)
        if self.optimizer == "nadams":
            optimizer = K.optimizers.Nadam(learning_rate=self.learningRate)

        self.model.compile(
            optimizer=optimizer,
            loss=lossFunction,
            loss_weights=lossWeights,
            metrics=self.getMetrics(),
            # run_eagerly=None,
            run_eagerly=True,  # for debug purpose! required for scorer ?
        )
        # for metric in self.validationMetrics:
        # m = scorer.getMetricScorer(metric)
        # self.model.addMetric(m, name='activation_mean')
        # if metric == ["acc", "accuracy"]:
        #     metric = "acc"
        # elif metric == "kappa":
        #     metric = tfa.metrics.CohenKappa(num_classes=self.numClasses)
        # elif metric == "auprc":
        #     metric = tf.keras.metrics.AUC(name="AUPRC", curve="PR")
        # elif metric == "auroc":
        #     metric = tf.keras.metrics.AUC(name="AUROC", curve="ROC")
        # elif metric == "meanSquared":
        #     metric = tf.keras.metrics.MeanSquaredError(name="meanSquared")
        # self.model.compile(
        #     optimizer="adam",
        #     loss="categorical_crossentropy",
        #     # loss_weights=lossWeights,
        #     metrics=["accuracy"],
        #     run_eagerly=None,  # for debug purpose!
        # )

    def summary(self):
        total_params = tf.reduce_sum([tf.reduce_prod(v.shape) for v in self.model.trainable_weights])
        self.log("Total trainable Parameters: %i" % (total_params))
        self.parameter = int(total_params)

        return self.model.summary()

    def __init__(self, options) -> None:
        super().__init__(options=options)

        if self.useGPU:
            physical_devices = tf.config.experimental.list_physical_devices("GPU")
            print("Num GPUs Available: ", len(physical_devices))
            if len(physical_devices) > 0:
                tf.config.experimental.set_memory_growth(physical_devices[0], True)

    def save(self, path):
        self.model.save_weights(path)

    def load(self, path):
        return self.model.load_weights(path)

    def loadState(self, state):
        return self.model.load_weights(state)

    def predict(self, input, get_likelihood=False, returnNumpy=True):
        batchSize = input.shape[0]
        additionalDimension = len(self.inputShape) == 3

        input = self.prepareX(input)

        if additionalDimension:
            input = input.reshape(-1, self.inputShape[0], self.inputShape[1], self.inputShape[2])
        input_tensor = tf.convert_to_tensor(input)
        output_tensor = self.model(input_tensor)
        out = output_tensor.numpy()

        if additionalDimension:
            out = out.reshape(1, out.shape[0] * out.shape[1], out.shape[2])

        predictions = self.mapOutputForPrediction(out)
        if self.oneHotDecoded:
            predictions = predictions.reshape(batchSize, -1, self.numClasses)

            if not get_likelihood:
                predictions = predictions.argmax(axis=2)

        # predict supposed to return Tensors: https://js.tensorflow.org/api/latest/#tf.Sequential.predict
        # but its allready numpy
        # if returnNumpy:
        #     predictions = predictions.numpy()

        return predictions

    def mapOutputForLoss(self, output, mask):
        return output.reshape(-1)



import gc
import os
import timeit

import numpy as np
import psutil
import torch
import torch.optim as optim
from tqdm import tqdm

from ..DataSet import DataSet, TrainingSet
from ..Model import Model
from ..scorer.ScorerTorch import ScorerTorch


class ModelTorchAdapter(Model):
    model: torch.nn.Module
    useGPU = torch.cuda.is_available()
    useTensorBoard = False
    saveOnlyWeights = True

    def _metrics(self, metric):
        return "val_" + metric["name"]

    def prepareX(self, x, validation=False):
        x = x.permute(0, 2, 1)
        return x

    def prepareY(self, y, validation=False):
        return y

    def init(self):
        self.weightTensors = None if self.classWeights is None else torch.from_numpy(np.array(self.classWeights))
        if self.useGPU and self.weightTensors is not None:
            self.weightTensors = self.weightTensors.cuda()

    def prepareDataAdapter(self, datasetOrTuple, validation=False):
        x, y = datasetOrTuple

        if not torch.is_tensor(x):
            x = torch.tensor(x)
        if not torch.is_tensor(y):
            y = torch.tensor(y)

        dataset = DataSet(x, y)

        if self.useGPU:
            dataset.x = dataset.x.cuda()
            dataset.y = dataset.y.cuda()

        return self.prepareData(dataset, validation)

    def remapLabels(self, categorizedArray):
        ones = torch.sparse.torch.eye(len(self.numClasses))
        return ones.index_select(0, categorizedArray)

    def evalValidation(self, validationData):
        model = self.model
        model.eval()
        s = ScorerTorch(self.numClasses, trace=True)
        # s = ScorerTorchEvent(self.numClasses, trace=True) if self.useEventScorer else ScorerTorch(self.numClasses, trace=True)
        s.metrics = self.validationMetrics
        s.trace = True
        s.ignoreClasses = [self.ignoreClassIndex] if self.ignoreClassIndex is not None else []

        validationData.transformToCategorical = 0
        batchCount = len(validationData)
        processList = tqdm(range(batchCount), disable=(not self.showProgress))
        processList.set_description("Validation")
        batchGenerator = validationData.generator()

        iter(validationData)
        for batchIndex in processList:
            validationBatch = batchGenerator.__next__()
            x, y = self.prepareDataAdapter(validationBatch, validation=True)

            # Run model
            output = model(x)
            batchPredictions = self.mapOutputForPrediction(output)

            if len(s.metrics) > 0:
                if self.oneHotDecoded:
                    y = y.reshape(-1, self.numClasses)
                    batchPredictions = batchPredictions.reshape(-1, self.numClasses)
                else:
                    y = y.reshape(-1)
                    batchPredictions = batchPredictions.reshape(-1)

                results = s.score(y, batchPredictions, trace=True)

                processList.set_postfix({m: results[m] for m in s.metrics})

            del batchPredictions
            del x
            del y
            del output
            gc.collect()

        if len(s.metrics) > 0:
            results = s.scoreAllRecords()

            metrics = {m: results[m] for m in s.metrics}
            justPrint = []

            if "confusion" in s.results:
                justPrint.append(s.results["confusion"])

            return metrics, justPrint

        return None, None

    def prepareTargetsForLoss(self, targets, oneHotDecoded=None):
        oneHotDecoded = self.oneHotDecoded if oneHotDecoded is None else oneHotDecoded
        mask = None
        if oneHotDecoded:
            targets = targets.reshape(-1, self.numClasses)
            if self.ignoreClassIndex is not None:
                mask = targets.eq(0).sum(-1) < self.numClasses
                targets = targets[mask]
        else:
            targets = targets.reshape(-1)

            if self.ignoreClassIndex is not None:
                mask = targets != self.ignoreClassIndex
                targets = targets[mask]

        return targets.float(), mask

    def train(self, dataset: TrainingSet):

        if ModelTorchAdapter.useTensorBoard:
            from torch.utils.tensorboard import SummaryWriter

            self.summarywriter = SummaryWriter()

            for (f, title) in self.startFigures:
                self.summarywriter.add_figure(title, f)
        else:
            self.summarywriter = None
        self.cleanCsv()

        metricDefinitions = {
            # init value, useAsBest, biggerIsBetter
            "AUC": [0.5, False, True],
            "AP": [0.0, True, True],
            "accuracy": [0.0, False, True],
            "kappa": [-1, True, True],
            "CSI": [0, True, True],
            "loss": [np.inf, True, False],
            "meanSquared": [np.inf, True, False],
            "SNR": [np.inf, False, False],
            "confusion": [np.inf, False, False],
            "auroc": [0.0, False, True],
            "auprc": [0.0, True, True],
            "eventCountDiff": [np.inf, False, False],
            "positiveErrorRatio": [np.inf, True, False],
        }

        model = self.model
        globalBestMetric = self.validationMetrics[0]

        self.log("LR: %s" % self.learningRate)
        if self.optimizer == "adams":
            decay = 0 if self.learningRateDecay is None else self.learningRateDecay
            self.optimizer = optim.Adam(model.parameters(), lr=self.learningRate, weight_decay=decay)
        if self.optimizer == "nadams":
            decay = 0 if self.learningRateDecay is None else self.learningRateDecay
            torch.optim.NAdam(model.parameters(), lr=self.learningRate, momentum_decay=decay)
        elif self.optimizer == "adas":
            from teleschlafmedizin.model.adapter.optimizer.Adas import Adas

            self.optimizer = Adas(model.parameters(), lr=self.learningRate)
        elif type(self.optimizer) == str:
            raise Exception("optimizer %s is currently not supported for pytorch implementation" % self.optimizer)

        model = self.model
        lossCriterion = self.getLossFunction()
        i_epoch = self.startEpoch
        notImprovedSince = 0
        lastImprovdModel = None

        stopAfterNotImproving = 0 if self.stopAfterNotImproving is None else self.stopAfterNotImproving

        while self.maxEpochs is None or i_epoch < self.maxEpochs:
            # Put in train mode
            trainingStartTime = timeit.default_timer()

            model.train(True)
            runningStats = {"loss": 0.0}
            batchesPerEpoch = len(dataset.trainingData)
            processList = tqdm(range(batchesPerEpoch), disable=(not self.showProgress))
            processList.set_description("EPOCH {}".format(i_epoch))
            batchGenerator = dataset.trainingData.generator()

            iter(dataset.trainingData)
            for batchIndex in processList:
                trainBatch = next(batchGenerator)
                batchFeats, targs = self.prepareDataAdapter(trainBatch)
                targs, mask = self.prepareTargetsForLoss(targs)

                self.optimizer.zero_grad()

                # check if there are any targets left
                if len(targs) == 0:
                    continue

                # for batchFeat in batchFeats:
                output = model(batchFeats)
                output = self.mapOutputForLoss(output, mask)
                loss = lossCriterion(output, targs)
                ownStats = hasattr(lossCriterion, "stats")

                if ownStats:
                    processList.set_postfix(ordered_dict=lossCriterion.stats)
                    for stat, value in lossCriterion.stats.items():
                        if stat not in runningStats:
                            runningStats[stat] = value
                        else:
                            runningStats[stat] += value

                # Backpropagation
                loss.backward()
                # torch.nn.utils.clip_grad()
                self.optimizer.step()

                # Perform one optimization step
                currentBatchLoss = loss.data.cpu().numpy()
                if np.isnan(currentBatchLoss):
                    model(batchFeats)
                    lossCriterion(output, targs)
                    raise Exception("batch loss should not be a number")

                runningStats["loss"] += currentBatchLoss

                del output
                del targs
                del loss
                gc.collect()
                currentCount = processList.n + 1
                processList.set_postfix(ordered_dict={n: v / currentCount for n, v in runningStats.items()})

            runningStats = {n: v / batchesPerEpoch for n, v in runningStats.items()}

            i_epoch += 1
            trainingEndTime = timeit.default_timer()

            # Get validation accuracy
            metricsValues, justPrint = self.evalValidation(dataset.validationData)

            metricStrings = []
            metricDiffStrings = []
            metricValuetrings = []
            improved = False
            modelId = "checkpointModel_%i_" % i_epoch

            for metricName, metricVal in metricsValues.items():
                bestValue, useAsBest, biggerIsBetter = metricDefinitions[metricName]
                diff = metricVal - bestValue
                metricStrings.append(metricName + ": " + "{:.3f}".format(metricVal) + " [best: {:.3f}]".format(bestValue))
                metricDiffStrings.append(metricName + ": " + "{:.3f}".format(diff))
                metricValuetrings.append("{:.3f}".format(metricVal))

                if self.summarywriter is not None:
                    self.summarywriter.add_scalar(metricName, metricVal, global_step=self.fullEpochs)

                isBigger = metricVal > bestValue

                if (biggerIsBetter and isBigger) or (not biggerIsBetter and not isBigger):
                    metricDefinitions[metricName][0] = metricVal
                    if useAsBest:
                        improved = True

            validationEndTime = timeit.default_timer()

            self.log(
                "Validation-Epoch Number: "
                + str(i_epoch)
                + "  Training Time: "
                + str(trainingEndTime - trainingStartTime)
                + "  Validation Time: "
                + str(validationEndTime - trainingEndTime)
            )
            for p in justPrint:
                self.log(p)

            trainingStats = " | ".join(["%s:%s" % (n, v) for n, v in runningStats.items()])
            self.log("Training Stats: %s " % (trainingStats))
            self.log(" ".join(metricStrings))
            # acc_train = correct / total
            # self.summarywriter.add_scalar("train_loss", runningLoss, global_step=self.fullEpochs)
            # self.summarywriter.add_scalar("train_acc", acc_train, global_step=self.fullEpochs)

            process = psutil.Process(os.getpid())
            self.log("memory usage: %sM" % (process.memory_info().rss / 1024 / 1024))

            csvRow = {
                "epoch": i_epoch,
            }
            csvRow.update(runningStats)

            for metricName in metricsValues:
                csvRow["val_%s" % metricName] = metricsValues[metricName]

            self.addCsvRow(csvRow)
            self.bestMetric = max(self.bestMetric, metricsValues[globalBestMetric])

            # If the validation accuracy improves, print out training and validation accuracy values and checkpoint the model
            if improved:
                self.log("Model Improved: " + " ".join(metricDiffStrings))
                f = open(
                    self.getModelPath() + "/" + modelId + "_".join(metricValuetrings) + ".pkl",
                    "wb",
                )
                # self.trigger("modelImproved", model)
                torch.save(model.state_dict() if ModelTorchAdapter.saveOnlyWeights else model, f)
                f.close()
                notImprovedSince = 0
                lastImprovdModel = model
            else:
                notImprovedSince += 1
                self.log("Model not improving since %i epochs" % (notImprovedSince))

            if stopAfterNotImproving > 0 and notImprovedSince >= stopAfterNotImproving:
                break

        self.csvClose()
        self.fullEpochs = i_epoch
        return lastImprovdModel

    def getModelPath(self):
        return self.logPath

    def build(self):

        torchSeed = 2
        torch.manual_seed(torchSeed)

        if self.useGPU:
            torch.cuda.manual_seed(torchSeed)
            self.model.cuda()

    def summary(self):
        pytorch_total_params = sum(p.numel() for p in self.model.parameters() if p.requires_grad)
        self.log("Total trainable Parameters: %i" % (pytorch_total_params))
        self.parameter = pytorch_total_params

        return str(self.model)

    def cleanUp(self):
        if self.useGPU:
            torch.cuda.empty_cache()

    def save(self, path):
        torch.save(self.model.state_dict(), path)

    def load(self, path):
        return torch.load(path, map_location=torch.device("cuda" if torch.cuda.is_available() else "cpu"))

    def loadState(self, state):
        if isinstance(state, torch.nn.Module):
            state = state.state_dict()
        return self.model.load_state_dict(state)

    def mapOutputForLoss(self, output, mask=None):
        output = output.reshape(-1, self.numClasses) if self.oneHotDecoded else output.flatten()
        return output[mask] if mask is not None else output

    def predict(self, input, get_likelihood=False, returnNumpy=True):
        with torch.no_grad():
            input = torch.from_numpy(input)
            batchSize, _, _ = input.shape

            if self.useGPU:
                input = input.cuda()

            input = self.prepareX(input)
            model = self.getModelEval()
            out = model(input)

            predictions = self.mapOutputForPrediction(out)

            if self.numClasses > 0 and self.oneHotDecoded:
                predictions = predictions.reshape(batchSize, -1, self.numClasses)

                if not get_likelihood:
                    predictions = torch.argmax(predictions, dim=2)

            if returnNumpy:
                predictions = predictions.detach().cpu().numpy()

            return predictions



from .Model import Model
from .ModelManager import ModelManager
from .DataSet import (
    DataSet,
    TrainingSet,
    TrainingSetLoader,
    DatasetWrap,
    DatasetWrapXY,
    DatasetWrapBatchGenerator,
    DatasetWrapItemGenerator,
)
from .scorer.Scorer import Scorer

from .AugmentationDataLoaderXY import AugmentationDataset, AugmentationDataLoaderXY
from .DataAugmentation import DataAugmentation
from .FeatureExtraction import FeatureExtraction

# import .adapter



import csv
from abc import ABC, abstractmethod
from pathlib import Path

from pyPhases.util.Optionizable import Optionizable

from .DataSet import DataSet


class MetricsDoesNotExist(Exception):
    pass


class Model(ABC, Optionizable):
    metrics = {
        "acc": {"name": "acc", "type": "max"},
        "binary_accuracy": {"name": "binary_accuracy", "type": "max"},
        "accuracy": {"name": "accuracy", "type": "max"},
        "kappa": {"name": "kappa", "type": "max"},
        "loss": {"name": "loss", "type": "min"},
    }
    optimizers = ["adams"]
    useGPU = True

    def __init__(self, options={}) -> None:
        super().__init__(options)
        self.logPath = "logs"
        self.csvLogFile = "log.csv"
        # segmentlength, channelSize
        self.inputShape = None
        self.numClasses = None
        self.classWeights = None

        self.batchSize = None
        self.maxEpochs = 100
        self.stopAfterNotImproving = None
        self.learningRate = 0.001
        self.learningRateDecay = None

        self.classNames = []
        self.monitorMetrics = ["accuracy"]
        self.optimizer = "adams"
        self.validationMetrics = ["kappa", "accuracy"]

        self.model = None
        self.modelEval = None
        self.modelDebug = None
        self.parameter = None
        self.bestMetric = 0
        self.fullEpochs = 0
        self.validationEvery = None
        self.ignoreClassIndex = -1

        self.startFigures = []
        self.showProgress = True
        self.startEpoch = 0
        self.useEventScorer = False
        self.predictionType = "classification"

        self.oneHotDecoded = False

    def getCsvPath(self):
        return self.logPath + "/" + self.csvLogFile

    def cleanCsv(self):
        parent = Path(self.getCsvPath()).parent
        if not parent.exists():
            Path(parent).mkdir(parents=True, exist_ok=True)
        if Path(self.getCsvPath()).exists():
            Path(self.getCsvPath()).unlink()

    def addCsvRow(self, row: dict):
        parent = Path(self.getCsvPath()).parent
        if not parent.exists():
            Path(parent).mkdir(parents=True, exist_ok=True)

        if not Path(self.getCsvPath()).exists():
            self.csvFileHanlder = open(self.getCsvPath(), "w+", newline="")
            self.writer = csv.writer(self.csvFileHanlder)
            self.writer.writerow(row.keys())

        self.writer.writerow(row.values())
        self.csvFileHanlder.flush()

    def csvClose(self):
        self.csvFileHanlder.close()

    def getMetric(self, name):
        if name not in Model.metrics:
            raise MetricsDoesNotExist("The metrics with the name %s is not defined" % name)
        return Model.metrics[name]

    def init(self):
        pass

    def define(self):
        pass

    def prepareData(self, dataset: DataSet, validation=False):
        x, y = dataset
        return self.prepareX(x), self.prepareY(y)

    def prepareX(self, x, validation=False):
        return x

    def prepareY(self, y, validation=False):
        return y

    def prepareRecordData(self, dataset: DataSet):
        shapeY = dataset.y.shape
        dataset.x = dataset.x.reshape(-1, self.inputShape[0], self.inputShape[1])
        dataset.y = dataset.y.reshape(shapeY[0] * shapeY[1], shapeY[2])
        return dataset

    def beforeTrain(self, dataset):
        pass

    def model(self):
        pass

    def summary(self):
        self.logWarning("No summary implemented in the model adapter")

    def getLossFunction(self):
        self.logError("No loss function was specified in the model adapter!")

    def getLossWeights(self):
        return None

    def getModelEval(self):
        model = self.model if self.modelEval is None else self.modelEval
        model.eval()
        return model

    def getModelDebug(self):
        return self.model if self.modelDebug is None else self.modelDebug

    def mapOutput(self, outputData):
        return outputData

    def mapPrediction(self, output):
        return output

    def mapOutputForPrediction(self, output, mask=None):
        # batchsize, predictionsCounts, classcount
        return output

    def mapOutputForLoss(self, output, mask=None):
        return output

    def cleanUp(self):
        return

    def save(self, path):
        return

    @abstractmethod
    def train(self):
        raise MetricsDoesNotExist("required train method is not implemented in the model adapter!")

    @abstractmethod
    def predict(self, input, get_likelihood=False, returnNumpy=True):
        raise MetricsDoesNotExist("required predict method is not implemented in the model adapter!")



import numpy as np
from numpy.random import default_rng
from pyPhases import classLogger


@classLogger
class DataAugmentation:
    def __init__(self, config, splitName) -> None:
        self.segmentAugmentation = config["segmentAugmentation"]
        self.recordAugmentation = config["recordAugmentation"]
        self.config = config
        self.splitName = splitName

    def step(self, stepname, X, Y, **options):
        # check if augmentation step exist

        if hasattr(self, stepname):
            # call method
            return getattr(self, stepname)(X, Y, **options)
        else:
            raise Exception(f"DataAugmentation {stepname} not found")

    def __call__(self, Segment, config):
        X, Y = Segment
        return self.augmentByConfig(X, Y, config)

    def augmentSegment(self, X, Y):
        return self.augmentByConfig(X, Y, self.segmentAugmentation)

    def augmentRecord(self, X, Y):
        return self.augmentByConfig(X, Y, self.recordAugmentation)

    def augmentByConfig(self, X, Y, config):
        X = np.expand_dims(X, axis=0)
        Y = np.expand_dims(Y, axis=0)

        for c in config:
            X, Y = self.loadFromConfig(c, X, Y, self.splitName)

        return X, Y

    def loadFromConfig(self, config, X, Y, splitName):
        config = config.copy()
        name = config["name"]
        ignoreChannels = config["ignoreChannels"] if "ignoreChannels" in config else None
        del config["name"]

        if "trainingOnly" in config:
            if config["trainingOnly"] and splitName != "training":
                return X, Y
            del config["trainingOnly"]

        # remove ignored channels for augmentation
        if ignoreChannels is not None:
            ignored = X[:, :, ignoreChannels]
            X = np.delete(X, ignoreChannels, axis=2)
            del config["ignoreChannels"]

        X, Y = self.step(name, X, Y, **config)

        # add ignored channels back
        if ignoreChannels is not None:
            X = np.insert(X, ignoreChannels, ignored, axis=2)

        return X, Y



import numpy as np
from sklearn.metrics import confusion_matrix

# from .model.EventManager import EventManager

class Scorer:
    title = "Segmentbasiert"
    cmScaleByClass = 4

    metricNameMap = {}

    def getMetricName(self, metricName):
        if metricName in Scorer.metricNameMap:
            metricName = Scorer.metricNameMap[metricName]
        return metricName

    def beforeLoss(self, x, y):
        """can be overwritten to tranform for a required format for the loss function"""
        return x, y

    def __init__(self, numClasses=None, classNames=None, trace=False, threshold=None) -> None:
        if numClasses is None and classNames is None:
            raise Exception("number of classes or list of classnames are required for scoring")

        self.metrics = ["kappa", "accuracy"]
        self.ignoreClasses = []
        self.numClasses = numClasses if classNames is None else len(classNames)
        self.classNames = [i for i in range(numClasses)] if classNames is None else classNames
        self.recordResult = {}
        self.recordIndex = 0
        self.trace = trace
        self.possibleValues = None
        self.lossF = None
        self.addEvents = False
        self.mapping = None
        self.threshold = threshold
        
    def prepareInput(self, inputArray):
        """can be overwritten to tranform tensors to numpy"""
        return inputArray

    def preparePrediction(self, inputArray):
        # for classification task the shape can be (recordcount?, predictions, classCount)
        isOneHotEncode = inputArray.shape[-1] == self.numClasses
        return inputArray.reshape(-1, self.numClasses) if isOneHotEncode else inputArray.reshape(-1)

    def prepareTruth(self, inputArray):
        """flatten and squeeze the truth array"""
        isOneHotEncode = inputArray.shape[-1] == self.numClasses

        if isOneHotEncode:
            maskedIgnored = inputArray.sum(axis=-1) == 0
            inputArray = np.argmax(inputArray, axis=-1)
            inputArray[maskedIgnored] = -1

        return inputArray.reshape(-1)

    def score(self, truth, prediction, recordName=None, trace=True):
        prediction = self.prepareInput(prediction)
        truth = self.prepareInput(truth)
        prediction = self.preparePrediction(prediction)
        truth = self.prepareTruth(truth)

        truthOrg, predictionOrg = truth, prediction
        truth, prediction = self.maskedIgnoredValues(truth, prediction)

        metrics = self.scoreMetrics(truth, prediction)
        if self.trace and trace:
            if recordName is None:
                recordName = self.recordIndex

            self.recordResult[recordName] = self.recordEntry(truthOrg, predictionOrg, self.results, recordName)

            self.recordIndex += 1

        return metrics

    def recordEntry(self, truth, prediction, metrics, recordName=None):

        prediction = self.flattenPrediction(prediction)

        # if "confusion" in metrics and self.addEvents:
        #     events = self.getValidationEvents(truth, prediction, recordName=recordName)
        # else:
        #     events = None

        entry = {
            "truth": truth,
            "prediction": prediction,
            # "events": events,
        }
        entry.update(metrics)
        return entry

    def getRecords(self):

        return self.recordResult.values()

    def getRecordsAsDataframe(self):
        import pandas as pd

        return pd.DataFrame(
            self.getRecords(),
            index=list(self.recordResult.keys()),
        )

    def combineMetric(self, metricName):

        metricName = self.getMetricName(metricName)

        result = None
        if metricName in ["kappa", "accuracy"]:  # recalcuate from confusion
            self.combineMetric("confusion")
            self.scoreMetric("kappa", None, None)
        elif metricName == "confusion":  # combine all confusion
            result = np.array(
                [self.recordResult[id]["confusion"] for id in self.recordResult if len(self.recordResult[id]["confusion"]) > 0]
            ).sum(axis=0)
        elif metricName in ["auprc", "auroc"]:

            if self.numClasses != 2:
                raise Exception("Metric auprc, auroc only support binary classification at the moment")

            self.results["all_values"] = np.array([self.recordResult[id]["all_values"] for id in self.recordResult]).sum(axis=0)
            self.results["pos_values"] = np.array([self.recordResult[id]["pos_values"] for id in self.recordResult]).sum(axis=0)
            self.results["ign_values"] = np.array([self.recordResult[id]["ign_values"] for id in self.recordResult]).sum(axis=0)

            # Compute the histogram of probabilities
            # given the above
            self.results["neg_values"] = self.results["all_values"] - self.results["pos_values"] - self.results["ign_values"]
            auroc, auprc, _, _, _ = self._auc(self.results["pos_values"], self.results["neg_values"])
            self.results["auprc"] = np.sum(auprc)
            self.results["auroc"] = np.sum(auroc)
        elif metricName in [
            "meanSquared",
            "loss",
            "rootMeanSquared",
            "prd",
            "eventCountDiff",
            "countTruth",
            "countPred",
            "overlappingCount",
            "positiveErrorRatio",
            "fm",
            "CSI",
            "SNR",
        ]:  # mean over all records
            sum = np.sum([self.recordResult[id][metricName] for id in self.recordResult])
            self.results[metricName] = sum / len(self.recordResult)

        if result is not None:
            self.results[metricName] = result

        return self.results[metricName]

    def combineMetrics(self):
        self.results = {}
        returnMetrics = {}
        for metricName in self.metrics:
            returnMetrics[metricName] = self.combineMetric(metricName)

        return returnMetrics

    def scoreAllRecords(self):
        if self.trace is False:
            raise Exception("The scorer does not keep the scored record, set the scorer.trace to True to keep all the results")
        self.combineMetrics()
        return self.results

    def maskedIgnoredValues(self, truth, prediction):
        for ignore in self.ignoreClasses:
            keep = truth != ignore
            truth = truth[keep]
            prediction = prediction[keep]

        return truth, prediction

    def scoreMetrics(self, truth, prediction):
        self.results = {}
        returnMetrics = {}
        for metricName in self.metrics:
            returnMetrics[metricName] = self.scoreMetric(metricName, truth, prediction)

        return returnMetrics

    def isHotEncoded(self, y):
        return y.shape[-1] == self.numClasses

    def calculateKappaAccFromConfusion(self, confusion):
        # implementation from sklearn
        n_classes = confusion.shape[0]
        sum0 = np.sum(confusion, axis=0)
        sum1 = np.sum(confusion, axis=1)
        if np.sum(sum0) == 0:
            return 0, 0
        expected = np.outer(sum0, sum1) / np.sum(sum0)

        w_mat = np.ones([n_classes, n_classes], dtype=np.int32)
        w_mat.flat[:: n_classes + 1] = 0

        nonDiag = np.sum(w_mat * confusion)
        if nonDiag == 0:
            k = 1
            acc = 1
        else:
            k = 1 - (nonDiag / np.sum(w_mat * expected))
            acc = 1 - (nonDiag / np.sum(confusion))

        return k, acc

    def maskArgMax(self, a, threshold):
        masked = np.ma.MaskedArray(a, a < threshold)
        return np.ma.argmax(masked)

    def getClassLikelyhood(self, prediction):
        max = np.max(prediction, axis=1, keepdims=True)
        if self.isHotEncoded(prediction):
            e_x = np.exp(prediction - max)
            sum = np.sum(e_x, axis=1, keepdims=True)  # returns sum of each row and keeps same dims
            predictionSoftMax = e_x / sum
            classLikelyhood = predictionSoftMax[:, 1]
        else:
            classLikelyhood = max
        return classLikelyhood

    def getClassPrediction(self, prediction):
        likelyhood = self.getClassLikelyhood(prediction)
        prediction = np.full(prediction.shape[0], 0)
        prediction[likelyhood > self.threshold] = 1
        return prediction

    def flattenPrediction(self, prediction, binaryLikelyHood=False, threshold=None):
        if binaryLikelyHood and self.numClasses != 2:
            raise Exception("Selected metric only support binary classification at the moment")

        if threshold is not None:
            return self.getClassPrediction(prediction)

        if self.isHotEncoded(prediction):
            prediction = prediction.argmax(-1)
        else:
            threshhold = 0.5 if self.threshold is None else self.threshold
            prediction[prediction >= threshhold] = 1
            prediction[prediction < threshhold] = 0
        return prediction

    def getMetricScorer(self, metricName):
        def score(truth, prediction):
            prediction = self.prepareInput(prediction)
            truth = self.prepareInput(truth)
            prediction = self.preparePrediction(prediction)
            truth = self.prepareTruth(truth)
            self.results = {}
            return self.scoreMetric(metricName, truth, prediction)

        score.__name__ = metricName
        return score

    def scoreMetric(self, metricName, truth, prediction):

        metricName = self.getMetricName(metricName)

        if metricName in self.results:
            return self.results[metricName]

        result = None
        if metricName in ["kappa", "accuracy"]:
            confusion = self.scoreMetric("confusion", truth, prediction)
            k, acc = self.calculateKappaAccFromConfusion(confusion)
            self.results["accuracy"] = acc
            self.results["kappa"] = k
        elif metricName == "confusion":
            prediction = self.flattenPrediction(prediction, threshold=self.threshold)
            result = confusion_matrix(
                truth, prediction, labels=range(self.numClasses)
            )  # , labels=range(self.numClasses) # HPC comp
        elif metricName in ["auprc", "auroc"]:
            classLikelyhood = self.getClassLikelyhood(prediction)

            scale = 10 ** 3
            b = scale + 1
            r = (-0.5 / scale, 1.0 + 0.5 / scale)

            self.results["all_values"] = np.histogram(classLikelyhood, bins=b, range=r)[0]

            pred_pos = classLikelyhood[truth > 0]
            self.results["pos_values"] = np.histogram(pred_pos, bins=b, range=r)[0]

            # Compute the histogram of probabilities within unscored regions
            pred_ign = classLikelyhood[truth < 0]
            self.results["ign_values"] = ign_values = np.histogram(pred_ign, bins=b, range=r)[0]

            # Compute the histogram of probabilities in non-arousal regions,
            # given the above
            self.results["neg_values"] = self.results["all_values"] - self.results["pos_values"] - self.results["ign_values"]
            auroc, auprc, _, _, _ = self._auc(self.results["pos_values"], self.results["neg_values"])
            self.results["auprc"] = np.sum(auprc)
            self.results["auroc"] = np.sum(auroc)
        elif metricName in ["loss"]:
            truth, prediction = self.beforeLoss(truth, self.flattenPrediction(prediction))
            self.results["loss"] = self.lossF(truth, prediction)
        elif metricName in ["meanSquared", "rootMeanSquared", "prd"]:
            if len(prediction.shape) > 1:
                diag = np.eye(self.numClasses)
                truth = diag[truth]
            squared = np.square(truth - prediction)
            meanSquared = squared.mean() if len(truth) > 0 else 0
            if metricName == "meanSquared":  # MS
                self.results[metricName] = meanSquared
            elif metricName == "rootMeanSquared":  # RMS
                self.results[metricName] = np.sqrt(meanSquared)
            elif metricName == "prd":  # Percentage RMS Difference
                self.results[metricName] = 100 * np.sqrt(squared.sum() / (np.square(truth).sum()))
        elif metricName in ["countTruth"]:
            truthCount = np.sum(np.diff(truth) > 0)
            self.results[metricName] = truthCount
        elif metricName in ["countPred"]:
            classPrediction = self.flattenPrediction(prediction, threshold=self.threshold)
            predCount = np.sum(np.diff(classPrediction) > 0)
            self.results[metricName] = predCount
        elif metricName in ["eventCountDiff"]:
            classPrediction = self.flattenPrediction(prediction, threshold=self.threshold)
            predCount = np.sum(np.diff(classPrediction) > 0)
            truthCount = np.sum(np.diff(truth) > 0)
            self.results[metricName] = int(np.abs(predCount - truthCount))
        elif metricName in ["overlappingCount"]:
            classPrediction = self.flattenPrediction(prediction, threshold=self.threshold)
            overlap = classPrediction + truth
            count = np.diff(overlap == 2).sum() / 2  # only count every up change
            self.results[metricName] = count
        elif metricName in ["positiveErrorRatio"]:
            confusion = self.scoreMetric("confusion", truth, prediction)
            tp = confusion[-1][-1]
            tn = confusion[0][0]
            f = np.sum(confusion) - tp - tn
            self.results[metricName] = f / tp if tp > 0 else 0
        elif metricName in ["CSI"]:
            confusion = self.scoreMetric("confusion", truth, prediction)
            tp = confusion[-1][-1]
            tn = confusion[0][0]
            f = np.sum(confusion) - tp - tn
            self.results[metricName] = tp / (tp + f)
        elif metricName == "SNR":  # signal to noise ratio
            axis = 0
            ddof = 0
            a = np.asanyarray(prediction)
            m = a.mean(axis)
            sd = a.std(axis=axis, ddof=ddof)
            self.results[metricName] = np.where(sd == 0, 0, m / sd)
        elif metricName == "fm":  # Fowlkes–Mallows index
            confusion = self.scoreMetric("confusion", truth, prediction)
            (tn, fp), (fn, tp) = confusion

            tp = confusion[-1][-1]
            tn = confusion[0][0]
            ppv = tp / (tp + fp)
            tpr = tp / (tp + fn)

            self.results[metricName] = np.sqrt(ppv * tpr)
        if result is not None:
            self.results[metricName] = result

        return self.results[metricName]

    def getValuesFromClasses(self):
        if self.possibleValues is None:
            values = []
            for id in range(self.numClasses ** 2):
                trueValue, preditedValue = (id // self.numClasses, id % self.numClasses)
                values.append("%s as %s" % (self.classNames[trueValue], self.classNames[preditedValue]))
            self.possibleValues = values

        return self.possibleValues

    # def getValidationEvents(self, truth, prediction, recordName=None):

    #     validationSignal = self.numClasses * truth + prediction
    #     validationSignal[truth == -1] = -1

    #     values = self.getValuesFromClasses()

    #     em = EventManager(None)
    #     return em.getEventsFromSignal(validationSignal, values, owner=recordName)

    # def confusionMatrix(self, resultEntry, labels=None, title="confusion matrix"):
    #     confusion = resultEntry["confusion"]
    #     labels = self.classNames

    #     label_list = [labels[i] for i in range(len(labels))]
    #     df_cm = pd.DataFrame(confusion, index=label_list, columns=label_list)

    #     blue = cm.get_cmap("Blues", 256)
    #     darkencolor = blue(np.linspace(0.3, 1, 256))
    #     newcmp = ListedColormap(darkencolor, name="DarkenBlue")
    #     figSize = self.numClasses * Scorer.cmScaleByClass
    #     return pp_matrix(df_cm, title=title, figsize=[figSize, figSize], cmap=newcmp, cbar=False)

    # def confusionMatrixTrailor(self, resultEntry, labels=None, title="sub confusion matrix", rowsSplit=slice(0,2), colSplit=slice(None), colNames = None, rowNames=None):
    #     #remove NoneRow
    #     colNames = self.classNames if colNames is None else colNames
    #     rowNames = self.classNames if rowNames is None else rowNames
    #     confusion = resultEntry["confusion"][colSplit, rowsSplit]
    #     labelsCol = np.array(colNames)[colSplit]
    #     labelsRow = np.array(rowNames)[rowsSplit]
    #     df_cm = pd.DataFrame(confusion, index=labelsCol, columns=labelsRow)

    #     blue = cm.get_cmap("Blues", 256)
    #     darkencolor = blue(np.linspace(0.3, 1, 256))
    #     newcmp = ListedColormap(darkencolor, name="DarkenBlue")
    #     figSizeL = (len(labelsCol) + 1 ) * Scorer.cmScaleByClass
    #     figSizeH = (len(labelsRow) +  1) * Scorer.cmScaleByClass
    #     return pp_matrix(df_cm, title=title, figsize=[figSizeL, figSizeH], cmap=newcmp, cbar=False)

    # def prCurve(self, resultEntry, title="ROC"):
    #     _, _, precision, _, recall = self._auc(resultEntry["pos_values"], resultEntry["neg_values"])
    #     fig = plt.figure(figsize=(6, 6))
    #     lw = 2
    #     plt.plot(
    #         recall,
    #         precision,
    #         color="darkorange",
    #         lw=lw,
    #         label=title,
    #     )
    #     plt.xlim([0.0, 1.0])
    #     plt.ylim([0.0, 1.05])
    #     plt.xlabel("Recall")
    #     plt.ylabel("Precision")
    #     plt.title("Precision/Recall Curve")
    #     plt.legend(loc="lower right")

    #     return fig

    # def rocCurve(self, resultEntry, title="ROC"):
    #     _, _, tpr, fpr, _ = self._auc(resultEntry["pos_values"], resultEntry["neg_values"])

    #     fig = plt.figure(figsize=(6, 6))
    #     lw = 2
    #     plt.plot(
    #         fpr,
    #         tpr,
    #         color="darkorange",
    #         lw=lw,
    #         label=title,
    #     )
    #     plt.plot([0, 1], [0, 1], color="navy", lw=lw, linestyle="--")
    #     plt.xlim([0.0, 1.0])
    #     plt.ylim([0.0, 1.05])
    #     plt.xlabel("False Positive Rate")
    #     plt.ylabel("True Positive Rate")
    #     plt.title("Receiver operating characteristic")
    #     plt.legend(loc="lower right")
    #     return fig

    def _auc(self, pos_values, neg_values):
        # Calculate areas under the ROC and PR curves by iterating
        # over the possible threshold values.

        # At the minimum threshold value, all samples are classified as
        # positive, and thus TPR = 1 and TNR = 0.
        tp = np.sum(pos_values)
        fp = np.sum(neg_values)
        tn = fn = 0
        tpr = 1
        tnr = 0
        if tp == 0 or fp == 0:
            # If either class is empty, scores are undefined.
            return (
                float("nan"),
                float("nan"),
                float("nan"),
                float("nan"),
                float("nan"),
            )
        ppv = float(tp) / (tp + fp)
        auroc = []
        auprc = []
        tpr_arr = []
        fpr_arr = []
        precisions = []

        # As the threshold increases, TP decreases (and FN increases)
        # by pos_values[i], while TN increases (and FP decreases) by
        # neg_values[i].
        for (n_pos, n_neg) in zip(pos_values, neg_values):
            tp -= n_pos
            fn += n_pos
            fp -= n_neg
            tn += n_neg
            tpr_prev = tpr
            tnr_prev = tnr
            ppv_prev = ppv
            tpr = float(tp) / (tp + fn)  # recall
            tnr = float(tn) / (tn + fp)
            fpr = float(fp) / (fp + tn)

            if tp + fp > 0:
                ppv = float(tp) / (tp + fp)  # precision
            else:
                ppv = ppv_prev

            auroc.append((tpr_prev - tpr) * (tnr + tnr_prev) * 0.5)
            auprc.append((tpr_prev - tpr) * ppv_prev)

            tpr_arr.append(tpr)
            fpr_arr.append(fpr)
            precisions.append(ppv)
        return (auroc, auprc, tpr_arr, fpr_arr, precisions)



from pyPhases import PluginAdapter, Project

from pyPhasesML.ModelManager import ModelManager

from .exporter.ModelExporter import ModelExporter


class Plugin(PluginAdapter):
    def __init__(self, project: Project, options=None):
        super().__init__(project, options)
        self.project.systemExporter["ModelExporter"] = "pyPhasesML"
        self.project.systemExporter["MemmapRecordExporter"] = "pyPhasesML"
        
    def initPlugin(self):
        # reload the model everytime the config changed
        def updateModel(changedValue):
            ModelManager.loadModel(self.project)

        self.project.on("configChanged", updateModel)

        self.project.registerExporter(ModelExporter())



import torch
from .Scorer import Scorer


class ScorerTorch(Scorer):
    def beforeLoss(self, x, y):
        return torch.from_numpy(x), torch.from_numpy(y)

    def prepareInput(self, inputArray):
        if isinstance(inputArray, torch.Tensor):
            inputArray = inputArray.detach().cpu().numpy()
        return inputArray



import tensorflow as tf

from .Scorer import Scorer


class ScorerTF(Scorer):
    def prepareInput(self, inputArray):
        if tf.is_tensor(inputArray):
            inputArray = inputArray.numpy()
            
        return inputArray


from pyPhases import classLogger
from pyPhasesRecordloader import RecordSignal, Signal


@classLogger
class FeatureExtraction:

    def step(self, stepname, recordSignal, **options):
        if hasattr(self, stepname):
            # call method
            return getattr(self, stepname)(recordSignal, **options)
        else:
            raise Exception(f"FeatureExtraction '{stepname}' not found")

    def __call__(self, recordSignal, config=None):
        return self.extractChannelsByConfig(recordSignal, config)

    def extractChannelsByConfig(self, recordSignal, config):

        for c in config:
            self.extractChannelByConfig(recordSignal, c)

    def addSingle(self, name, recordSignal: RecordSignal, signalOrArray):
        if not isinstance(signalOrArray, Signal):
            signalOrArray = Signal(name, signalOrArray, frequency=recordSignal.targetFrequency)
        signalOrArray.name = name
        recordSignal.addSignal(signalOrArray, name)

    def extractChannelByConfig(self, recordSignal, config):
        config = config.copy()
        name = config["name"]
        channels = config["channels"]
        del config["name"]
        del config["channels"]

        ret = self.step(name, recordSignal, **config)

        if len(channels) > 1:
            for index, signal in enumerate(ret):
                self.addSingle(channels[index], recordSignal, signal)
        elif ret is not None:
            self.addSingle(channels[0], recordSignal, ret)
        
